(*
Copyright (C) 2003-2004 Evgenii Rudnyi, http://Evgenii.Rudnyi.Ru

This software is a copyrighted work licensed under the terms, described
in the file "FREE_LICENSE". 
*)

(* 
	Computational functions to estimate thrust and impulse under
	the ideal rocket theory.
	
	See Specification Manual for details: Eqs (4) - (10)
	There are some more equations for subsonic flow
	when pt/p1 < critialRatio or mflow < mflowCritical
	then
		mflowCritical = p1*At*Sqrt[k/R/T1*((k + 1)/2)^((1 + k)/(1 - k))]
		mflow = A2*p1*Sqrt[2*k/(k-1)/R/T1*(pratio^(2/k) - pratio^((k+1)/k))]
		p2 = pex
		pratio = pex/p2
	
	Notation:
		p1, p2 - pressure at inlet and outlet (Pa)
		v1, v2 - velocity at inlet and outlet (v1 is evaluated at x = l)
		k - Cp/Cv ratio
		R - specific gas constant (R/M) (J/(kg K))
		pex - external pressure (Pa)
		At, A2 - area at throat and outlet (m^2)
		A(x) - function how area depend on distance in the nozzle
			x changes from zero to l with the burning front
			direction of the axis toward back of the nozzle
		mflow - mass flow (kg/s)
		rofuel - density of the fuel (kg/m^3)
		T1 - temperature of the flame (K)
		r(p) - burning rate (m/s), depends on pressure
		V0 - initial volume in the chamber (% of the total value) can not be zero
		p0 = pex - initial p1 in the chamber (Pa) 
		F - thrust
		Imp - impulse
		pratio = p2/p1
		
		All output parameters except v2, criticalRatio, pratio 
		are functions in x
			(InterpolationFunction objects)
			
		A shock wave within the nozzle is partly included. It is necessary to do it more nicely.

		An assumption is that S(x) is monotonically non-decreasing
*)

(* output	{p1, p2, t, v1, mflow, F, Imp, v2, xCr1, xCr2} *)

thrust[k_, R_, pex_, A_, l_, At_, A2_, rofuel_, T1_, r_, V0_] :=
Module[{mflowMul, mflowSuper, mflowSub, mflowSubGrad, pmax, pcr, pcr2, pin,  Vini, V, eq, p1, t, p, x, pr, pratio, fun, p2, v1, v2, mflow, F, Imp, xCr1, xCr2},

	(* multiply to compute mass flow: mflow = p1*mflowMul *)
	mflowSuper = At*Sqrt[k/(R*T1)*((2./(k + 1))^((k+1)/(k-1)))];
	mflowSub[pin_] := A2*Re[Sqrt[2*k/(k-1)/R/T1*((pex/pin)^(2/k) - (pex/pin)^((k+1)/k))]];
	
	mflowSubGrad = ( 
		(A2*k*((-2*pex*(pex/#)^(-1 + 2/k))/
			(k*#^2) + ((1 + k)*pex*
			 (pex/#)^(-1 + (1 + k)/k))/
			(k*#^2)))/(Sqrt[2]*(-1 + k)*R*
		Sqrt[(k*((pex/#)^(2/k) - 
				(pex/#)^((1 + k)/k)))/
			((-1 + k)*R*T1)]*T1))&;
   
	mflowMul[pin_] = If[pex <= 0.01,
			mflowSuper
		,
			pmax = bisection[mflowSubGrad, pex*1.00001, pex*5];
			pmax = If [Head[pmax] =!= List, pmax, 5*pex];
			pcr = bisection[(mflowSub[#] - mflowSuper)&, pex, pmax];
			pcr = If [Head[pcr] =!= List, pcr, pex];
			If[pcr > pex,
				If[pin < pcr, mflowSub[pin], mflowSuper]
			,
				mflowSuper
			]
		];
		
	(* let us solve mass balance eq to determine p1(x): 
		Change of variable in Eq 10: t -> x : df/dt = df/dx r *)
	V = NDSolve[{V'[x] == A[x], V[0] == 0}, V, {x, 0, l}][[1,1,2]];
	V = Evaluate[V[#] + V0*V[l]] &;

	eq := A[x]*r[p[x]]*rofuel == r[p[x]]*(A[x]*p[x] + p'[x]*V[x])/R/T1 + p[x]*mflowMul[p[x]];
	p1 = NDSolve[{eq, p[0] == pex}, p, {x, 0, l}, MaxSteps->10000][[1,1,2]];
	t = NDSolve[{t'[x] == 1/r[p1[x]], t[0] == 0.}, t, {x, 0, l}, MaxStepSize->l/1000, MaxSteps->10000][[1,1,2]];
	
	xCr1 = If [pex <= 0.01, 0, bisection[(p1[#] - pcr)&, 0, l]];
	xCr1 = If [Head[xCr1] === List, 
		If[xCr1[[1]] < 0, l, 0]
	, 
		xCr1
	];
	
	eq = (At/A2 - ((k + 1)/2)^(1/(k - 1))*#^(1/k)*Sqrt[(k + 1)/(k - 1)*(1 - #^((k-1)/k))])&;
	pr = bisection[eq, 0, 0.6];
	If [!NumericQ[pr], Print["can not find presure ration ", pr]; Return[]];

	pcr2 = -(((-1 + k^2)*pex)/((1 + 2*k + k^2 - (4*k)/(pr^(-1))^((1 - k)/k))*pr));
	xCr2 = bisection[(p1[#] - pcr2)&, xCr1, l];
	xCr2 = If [Head[xCr2] === List, 
		If[xCr2[[1]] < 0, l, xCr1]
	, 
		xCr2
	];	
	
	p2 = Evaluate[If[# <= Evaluate[xCr2], Evaluate[pex], Evaluate[p1[#]*pr]]] &;
	v1 = Evaluate[mflowSuper*R*T1/A[#]] &;
	
	v2 = Evaluate[Which[
		# <= xCr1,
			Sqrt[2*k/(k - 1)*R*T1*(1 - (pex/p1[#])^((k - 1)/k))],
		# <= xCr2,
			Module[{v21, Mr2, vr, v22},
				v21 = Sqrt[2*k/(k - 1)*R*T1*(1 - (pex/p1[xCr1])^((k - 1)/k))];
				Mr2 = ((1/pr)^((k-1)/k) - 1)/(k - 1)*2;
				vr = (k - 1)/(k + 1)*(1 + 2/(k - 1)/Mr2);
				v22 = Sqrt[2*k/(k - 1)*R*T1*(1 - pr^((k - 1)/k))]*vr;
				v21+(v22-v21)/(xCr2-xCr1)*(#-xCr1)
			],
		True,
			Sqrt[2*k/(k - 1)*R*T1*(1 - pr^((k - 1)/k))]
	]] &;
	Off[FunctionInterpolation::"ncvb"];
	v2 = FunctionInterpolation[v2[x], {x, 0, l}, InterpolationPoints -> 1000];
	mflow = FunctionInterpolation[p1[x]*mflowMul[p1[x]], {x, 0, l}, InterpolationPoints -> 1000];
	
	F = Evaluate[mflow[#]*v2[#] + (p2[#] - pex)*A2] &;
	Imp = NDSolve[{Imp'[x] == F[x]/r[p1[x]], Imp[0] == 0}, Imp, {x, 0, l}][[1,1,2]];
	{p1, p2, t, v1, mflow, F, Imp, v2, xCr1, xCr2}
];

(*simple bisection - a < b*)
bisection[f_, a_, b_] := Module[{eps = 1*^-14, fl, fr, xl, xr},
	xl = a;
	fl = f[a]; 
	If [!NumericQ[fl], Print["not numeric function ", fl]; Return[]];
	If[fl == 0., Return[a]];
	xr = b;
	fr = f[b];
	If [!NumericQ[fr], Print["not numeric function ", fr]; Return[]];
	If[fr == 0., Return[b]];
	If [Sign[fl]*Sign[fr] > 0., Return[{Sign[fl]}]];
	While[Abs[xr - xl] > eps*Abs[xr + xl],
		xm = (xr + xl)/2.;
		fm = f[xm];
		If [!NumericQ[fm], Print["not numeric function ", fm]; Return[]];
		If[fm == 0., Return[xm]];
		If[Sign[fl]*Sign[fm] < 0.,
			xr = xm; fr = fm;
		,
			xl = xm; fl = fm;
		]
	];
	(xr + xl)/2
];
